@@ -1,4 +1,4 @@
-# Copyright (c) 2020 Sine Nomine Associates
+# Copyright (c) 2020-2021 Sine Nomine Associates
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -27,11 +27,13 @@ import configparser
 import crypt
 import datetime
 import fcntl
-import io
+import getpass
 import glob
+import io
 import json
 import logging
 import os
+import pprint
 import secrets
 import shlex
 import socket
@@ -42,14 +44,18 @@ import xml.etree.ElementTree
 import sh
 import libvirt
 
-from .config import SETTINGS, TEMPLATES, SCRIPTS
-
 log = logging.getLogger(__name__)
 
 # Environment variables
 libvirt_uri = os.environ.get('LIBVIRT_DEFAULT_URI', 'qemu:///session')
-xdg_config_home = os.path.expanduser(os.environ.get('XDG_CONFIG_HOME', '~/.config'))
-xdg_data_home = os.path.expanduser(os.environ.get('XDG_DATA_HOME', '~/.local/share'))
+
+virtup_config_home = os.path.expanduser(
+    os.environ.get('VIRTUP_CONFIG_HOME',
+    os.path.join(os.environ.get('XDG_CONFIG_HOME', '~/.config'), 'virt-up')))
+
+virtup_data_home = os.path.expanduser(
+    os.environ.get('VIRTUP_DATA_HOME',
+    os.path.join(os.environ.get('XDG_DATA_HOME', '~/.local/share'), 'virt-up')))
 
 # Helpers
 def rm_f(path):
@@ -98,45 +104,57 @@ libvirt.registerErrorHandler(f=_libvirt_callback, ctx=None)
 
 class Settings:
     """
-    Configuration settings read from the settings (ini) file
-    for the local libvirt settings and OS specific settings.
+    Configuration settings for a given template definition name.
     """
-    def __init__(self, template=None):
-        # Load local site settings.
-        self.settings = self._load('settings', SETTINGS)
-        self.site = self.settings.get('site', {})
-        self.pool = self.site.get('pool', 'default')
-        self.username = self.site.get('username', 'virt')
-        self.image_format = self.site.get('image-format', 'qcow2')
-        self.dns_domain = self.site.get('dns-domain', '')
-        self.address_source = self.site.get('address-source', 'agent')
-        self.template_playbook = self.site.get('template-playbook', '')
-        self.instance_playbook = self.site.get('instance-playbook', '')
+    def __init__(self, name):
+        """
+        Get template definition settings.
+        """
+        # Load common settings.
+        settings = self._load('settings.cfg')
+        common = settings.get('common', {})
 
-        # Load template definitions.
-        self.templates = self._load('templates', TEMPLATES)
-        if template is None:
-            self.template = {}
-            self.template_name = None
-            self.ov_version = None
-            self.os_variant = None
-        else:
-            if not template in self.templates:
-                raise LookupError(f"Template '{template}' not found in settings.")
-            self.template = self.templates[template]
-            self.template_name = template
-            self.template_playbook = self.template.get('template-playbook')
-            self.instance_playbook = self.template.get('instance-playbook')
-            self.os_version = self.template.get('os-version')
-            self.os_variant = self.template.get('os-variant')
-            self.address_source = self.template.get('address-source', self.address_source)
+        # Load template specific settings.
+        templates = self._load('templates.d/*.cfg')
+        if not name in templates:
+            raise LookupError(f"Template '{name}' not found in settings.")
 
-    def _load(self, name, defaults):
-        system_file = f'/etc/virt-up/{name}.cfg'
-        user_file = f'{xdg_config_home}/virt-up/{name}.cfg'
+        template = templates[name]
+        def get(option, default):
+            return template.get(option, common.get(option, default))
+
+        self.template_name = name
+        self.desc = get('desc', '')
+        self.os_version = get('os-version', '')
+        self.os_variant = get('os-variant', '')
+        self.arch = get('arch', '')
+        self.pool = get('pool', 'default')
+        self.user = get('user', get('username', getpass.getuser()))
+        self.image_format = get('image-format', 'qcow2')
+        self.memory = get('memory', 512)
+        self.vcpus = get('vcpus', 1)
+        self.graphics = get('graphics', 'none')
+        self.dns_domain = get('dns-domain', '')
+        self.address_source = get('address-source', 'agent')
+        self.virt_builder_args = shlex.split(get('virt-builder-args', ''))
+        self.virt_sysprep_args = shlex.split(get('virt-sysprep-args', ''))
+        self.virt_install_args = shlex.split(get('virt-install-args', ''))
+        self.cp_args = shlex.split(get('cp-args', ''))
+        self.template_playbook = get('template-playbook', '')
+        self.instance_playbook = get('instance-playbook', '')
+        log.debug("Settings: %s", pprint.pformat(vars(self)))
+
+    @classmethod
+    def _load(self, pattern):
+        """
+        Load settings from config files.
+        """
+        system_files = glob.glob(f'/etc/virt-up/{pattern}')
+        user_files = glob.glob(f'{virtup_config_home}/{pattern}')
         parser = configparser.ConfigParser()
-        parser.read_string(defaults)
-        parser.read([system_file, user_file])
+        filesread = parser.read(system_files + user_files)
+        for f in filesread:
+            log.debug(f"Read: {f}")
 
         # Convert to a regular dict and remove newlines.
         settings = {}
@@ -147,18 +165,13 @@ class Settings:
 
         return settings
 
-    def extra_args(self, command, **variables):
+    @classmethod
+    def all(cls):
         """
-        Extra command line arguments for the local hypervisor and the target
-        template definition.
+        Generator to list each template definition.
         """
-        args = []
-        for section in ('site', 'template'):
-            settings = getattr(self, section, {})
-            args_str = settings.get(f'{command}-args', '')
-            args_str = args_str.format(**variables)
-            args.extend(shlex.split(args_str))
-        return args
+        for name in cls._load('templates.d/*.cfg'):
+            yield Settings(name)
 
 class LockFile:
     """
@@ -231,7 +244,7 @@ class Creds:
         """
         Generate a ssh key pair for passwordless ssh login.
         """
-        ssh_identity = f'{xdg_data_home}/virt-up/sshkeys/{name}'
+        ssh_identity = f'{virtup_data_home}/sshkeys/{name}'
         if os.path.exists(ssh_identity):
             log.debug(f"SSH key file '{ssh_identity}' already exists.")
             if not os.path.exists(f'{ssh_identity}.pub'):
@@ -244,7 +257,7 @@ class Creds:
         return ssh_identity
 
 class MacAddresses:
-    filename = f'{xdg_data_home}/virt-up/macaddrs.json'
+    filename = f'{virtup_data_home}/macaddrs.json'
 
     """
     Saved instance mac addresses.
@@ -303,7 +316,7 @@ class Instance:
     """
     def __init__(self, name, meta=None):
         self.name = name
-        self.metafile = f'{xdg_data_home}/virt-up/instance/{name}.json'
+        self.metafile = f'{virtup_data_home}/instance/{name}.json'
         self.meta = {}
         self._mac = None
         self._disks = None
@@ -537,6 +550,18 @@ class Instance:
                 arp[mac] = address
         return arp
 
+    def _ping(self, address):
+        """
+        Verify the address is pingable.
+        """
+        ping = sh.Command('ping')
+        try:
+            ping('-c', 2, address)
+            return True
+        except sh.ErrorReturnCode as e:
+            log.debug(f"Unable to ping address '{address}'; ping code {e.exit_code}.")
+            return False
+
     def _address_from_arp(self):
         """
         Attempt to retreive the instance address from the arp cache.
@@ -546,7 +571,7 @@ class Instance:
         for retries in range(120, -1, -1):
             arp = self._arp_table()
             address = arp.get(mac)
-            if address:
+            if address and self._ping(address):
                 break
             if retries > 0:
                 suffix = 'ies' if retries > 1 else 'y'
@@ -568,7 +593,7 @@ class Instance:
                 ai = socket.getaddrinfo(hostname, 22, family=socket.AF_INET, proto=socket.IPPROTO_TCP)
                 if ai:
                     address = ai[0][4][0]
-                if address:
+                if address and self._ping(address):
                     break
             except:
                 pass
@@ -633,7 +658,7 @@ class Instance:
 
     @classmethod
     def all(cls):
-        pattern = f'{xdg_data_home}/virt-up/instance/*.json'
+        pattern = f'{virtup_data_home}/instance/*.json'
         for path in glob.glob(pattern):
             name = os.path.basename(path).replace('.json', '')
             yield Instance(name)
@@ -659,7 +684,7 @@ class Instance:
         Returns true if the instance already exists, that is
         the domain and metadata file both exist.
         """
-        metafile = f'{xdg_data_home}/virt-up/instance/{name}.json'
+        metafile = f'{virtup_data_home}/instance/{name}.json'
         return os.path.exists(metafile) and cls._domain_exists(name)
 
     @classmethod
@@ -671,10 +696,10 @@ class Instance:
               root_password=None,
               user=None,
               password=None,
-              memory=512,
+              memory=None,
               size=None,
-              vcpus=1,
-              graphics='none',
+              vcpus=None,
+              graphics=None,
               dns_domain=None,
               **kwargs):
         """
@@ -711,19 +736,9 @@ class Instance:
         if os.path.exists(image):
             raise FileExistsError(f"Image file '{image}' already exists.")
 
-        # Install virt-builder --run/--firstboot scripts.
-        for script in SCRIPTS:
-            path = os.path.join('/tmp/virt-up/scripts', script['name'])
-            if not os.path.exists(path):
-                if not os.path.exists('/tmp/virt-up/scripts'):
-                    os.makedirs('/tmp/virt-up/scripts')
-                with open(path, 'w') as f:
-                    f.write(script['contents'])
-                log.info('Wrote: %s', path)
-
         # Generate the ssh keys. Generate the passwords if not provided.
         if not user:
-            user = settings.username
+            user = settings.user
         root_creds = Creds('root', password=root_password)
         user_creds = Creds(user, password=password)
 
@@ -732,13 +747,19 @@ class Instance:
         password = crypt.crypt(user_creds.password, salt)
 
         # Setup virt-builder arguments.
+        if memory is None:
+            memory = settings.memory
+        if vcpus is None:
+            vcpus = settings.vcpus
+        if graphics is None:
+            graphics = settings.graphics
         if dns_domain is None:
             dns_domain = settings.dns_domain
         if dns_domain:
             hostname = f'{name}.{dns_domain}'
         else:
             hostname = name
-        extra_args = settings.extra_args('virt-builder')
+        extra_args = settings.virt_builder_args
         if size:
             extra_args.extend(['--size', size])
 
@@ -754,7 +775,7 @@ class Instance:
                 '--run-command', f"useradd -m -s /bin/bash -p '{password}' {user}",
                 '--ssh-inject', f'{user}:file:{user_creds.ssh_identity}.pub',
                 '--run-command', 'mkdir -p /etc/sudoers.d',
-                '--write',  f'/etc/sudoers.d/99-sna-devlab:{user} ALL=(ALL) NOPASSWD: ALL',
+                '--write',  f'/etc/sudoers.d/99-virt-up:{user} ALL=(ALL) NOPASSWD: ALL',
                 *extra_args)
 
         # Setup virt-install options. Reuse the last mac address for this
@@ -763,8 +784,7 @@ class Instance:
         mac = maddrs.lookup(name)
         if mac:
             optional_args.extend(['--mac', mac])
-        extra_args = settings.extra_args('virt-install')
-
+        extra_args = settings.virt_install_args
         with LockFile():
             log.info(f"Importing instance '{name}'.")
             virt_install(
@@ -857,7 +877,7 @@ class Instance:
             if settings.image_format == 'qcow2':
                 qemu_img.create('-f', 'qcow2', '-F', 'qcow2', '-b', source_image, target_image)
             else:
-                extra_args = settings.extra_args('cp')
+                extra_args = settings.cp_args
                 cp(*extra_args, source_image, target_image)
 
         # Setup virt-sysprep args.
@@ -869,7 +889,7 @@ class Instance:
             else:
                 hostname = target
 
-        extra_args = settings.extra_args('virt-sysprep')
+        extra_args = settings.virt_sysprep_args
 
         with LockFile():
             log.info(f"Preparing target image '{target_image}'.")
@@ -894,7 +914,7 @@ class Instance:
         if mac:
             optional_args.extend(['--mac', mac])
 
-        extra_args = settings.extra_args('virt-install')
+        extra_args = settings.virt_install_args
 
         with LockFile():
             log.info(f"Importing instance '{target}'.")
@@ -938,7 +958,7 @@ class Instance:
         """
         Create an ansible inventory file for the cloned instances.
         """
-        filename = f'{xdg_data_home}/virt-up/inventory.yaml'
+        filename = f'{virtup_data_home}/inventory.yaml'
         ssh_options = [
             ('ControlMaster', 'auto'),
             ('ControlPersist', '60s'),
@@ -1052,16 +1072,29 @@ class Instance:
         """
         Run an ansible playbook on this instance.
         """
-        inventory = f'{xdg_data_home}/virt-up/inventory.yaml'
+        inventory = f'{virtup_data_home}/inventory.yaml'
+        if not ansible:
+            log.error("Skipping playbook; 'ansible-playbook' command not found.")
+            return
+        # Search for the playbook.
+        searched = []
+        found = None
+        for p in (playbook,
+                  os.path.join(virtup_config_home, 'playbooks', playbook),
+                  os.path.join('/etc/virt-up/playbooks', playbook)):
+            searched.append(p)
+            if os.path.exists(p):
+                found = p
+                break
+        if not found:
+            searched = ' '.join(searched)
+            log.error(f"Skipping playbook; '{playbook}' file not found. Searched: {searched}.")
+            return
+        playbook = found
+        self.wait_for_port(22)
         address = self.address()
         if not address:
             log.warning(f"Skipping playbook; address for '{self.name}' is not available.")
             return
-        if not os.path.exists(playbook):
-            log.warning(f"Skipping playbook; '{playbook}' file not found.")
-            return
-        if not ansible:
-            log.warning("Skipping playbook; 'ansible-playbook' command not found.")
-            return
-        log.info(f"Running playbook '{playbook}'.")
+        log.info(f"Running playbook '{playbook}' on '{self.name}'.")
         ansible('-i', inventory, '--limit', self.name, playbook)
